package com.sun.xml.internal.ws.policy;

import java.util.Collection;
import com.sun.xml.internal.ws.policy.spi.PolicyAssertionValidator;
import java.util.LinkedList;
import com.sun.xml.internal.ws.policy.privateutil.LocalizationMessages;
import java.util.Iterator;
import com.sun.xml.internal.ws.policy.privateutil.PolicyLogger;

public class EffectiveAlternativeSelector
{
    private static final PolicyLogger LOGGER;
    
    public EffectiveAlternativeSelector() {
    }
    
    public static void doSelection(EffectivePolicyModifier modifier) throws PolicyException {
        AssertionValidationProcessor validationProcessor = AssertionValidationProcessor.getInstance();
        selectAlternatives(modifier, validationProcessor);
    }
    
    protected static void selectAlternatives(EffectivePolicyModifier modifier, AssertionValidationProcessor validationProcessor) throws PolicyException {
        PolicyMap map = modifier.getMap();
        for (PolicyMapKey mapKey : map.getAllServiceScopeKeys()) {
            Policy oldPolicy = map.getServiceEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForServiceScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
        for (PolicyMapKey mapKey : map.getAllEndpointScopeKeys()) {
            Policy oldPolicy = map.getEndpointEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForEndpointScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
        for (PolicyMapKey mapKey : map.getAllOperationScopeKeys()) {
            Policy oldPolicy = map.getOperationEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForOperationScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
        for (PolicyMapKey mapKey : map.getAllInputMessageScopeKeys()) {
            Policy oldPolicy = map.getInputMessageEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForInputMessageScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
        for (PolicyMapKey mapKey : map.getAllOutputMessageScopeKeys()) {
            Policy oldPolicy = map.getOutputMessageEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForOutputMessageScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
        for (PolicyMapKey mapKey : map.getAllFaultMessageScopeKeys()) {
            Policy oldPolicy = map.getFaultMessageEffectivePolicy(mapKey);
            modifier.setNewEffectivePolicyForFaultMessageScope(mapKey, selectBestAlternative(oldPolicy, validationProcessor));
        }
    }
    
    private static Policy selectBestAlternative(Policy policy, AssertionValidationProcessor validationProcessor) throws PolicyException {
        AssertionSet bestAlternative = null;
        EffectiveAlternativeSelector.AlternativeFitness bestAlternativeFitness = EffectiveAlternativeSelector.AlternativeFitness.UNEVALUATED;
        for (AssertionSet alternative : policy) {
            EffectiveAlternativeSelector.AlternativeFitness alternativeFitness = alternative.isEmpty() ? EffectiveAlternativeSelector.AlternativeFitness.SUPPORTED_EMPTY : EffectiveAlternativeSelector.AlternativeFitness.UNEVALUATED;
            for (PolicyAssertion assertion : alternative) {
                PolicyAssertionValidator.Fitness assertionFitness = validationProcessor.validateClientSide(assertion);
                switch (assertionFitness) {
                    case UNKNOWN:
                    case UNSUPPORTED:
                    case INVALID: {
                        EffectiveAlternativeSelector.LOGGER.warning(LocalizationMessages.WSP_0075_PROBLEMATIC_ASSERTION_STATE(assertion.getName(), assertionFitness));
                        break;
                    }
                }
                alternativeFitness = alternativeFitness.combine(assertionFitness);
            }
            if (bestAlternativeFitness.compareTo(alternativeFitness) < 0) {
                bestAlternative = alternative;
                bestAlternativeFitness = alternativeFitness;
            }
            if (bestAlternativeFitness == EffectiveAlternativeSelector.AlternativeFitness.SUPPORTED) {
                break;
            }
            continue;
        }
        switch (bestAlternativeFitness) {
            case INVALID: {
                throw (PolicyException)EffectiveAlternativeSelector.LOGGER.logSevereException(new PolicyException(LocalizationMessages.WSP_0053_INVALID_CLIENT_SIDE_ALTERNATIVE()));
            }
            case UNKNOWN:
            case UNSUPPORTED:
            case PARTIALLY_SUPPORTED: {
                EffectiveAlternativeSelector.LOGGER.warning(LocalizationMessages.WSP_0019_SUBOPTIMAL_ALTERNATIVE_SELECTED(bestAlternativeFitness));
                break;
            }
        }
        Collection<AssertionSet> alternativeSet = null;
        if (bestAlternative != null) {
            alternativeSet = new LinkedList();
            alternativeSet.add(bestAlternative);
        }
        return Policy.createPolicy(policy.getNamespaceVersion(), policy.getName(), policy.getId(), alternativeSet);
    }
    
    static {
        LOGGER = PolicyLogger.getLogger(EffectiveAlternativeSelector.class);
    }
    
    private enum AlternativeFitness
    {
        UNEVALUATED {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                switch (assertionFitness) {
                    case UNKNOWN: {
                        return EffectiveAlternativeSelector.AlternativeFitness.1.UNKNOWN;
                    }
                    case UNSUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.1.UNSUPPORTED;
                    }
                    case SUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.1.SUPPORTED;
                    }
                    case INVALID: {
                        return EffectiveAlternativeSelector.AlternativeFitness.1.INVALID;
                    }
                    default: {
                        return EffectiveAlternativeSelector.AlternativeFitness.1.UNEVALUATED;
                    }
                }
            }
        }, 
        INVALID {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                return EffectiveAlternativeSelector.AlternativeFitness.2.INVALID;
            }
        }, 
        UNKNOWN {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                switch (assertionFitness) {
                    case UNKNOWN: {
                        return EffectiveAlternativeSelector.AlternativeFitness.3.UNKNOWN;
                    }
                    case UNSUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.3.UNSUPPORTED;
                    }
                    case SUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.3.PARTIALLY_SUPPORTED;
                    }
                    case INVALID: {
                        return EffectiveAlternativeSelector.AlternativeFitness.3.INVALID;
                    }
                    default: {
                        return EffectiveAlternativeSelector.AlternativeFitness.3.UNEVALUATED;
                    }
                }
            }
        }, 
        UNSUPPORTED {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                switch (assertionFitness) {
                    case UNKNOWN:
                    case UNSUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.4.UNSUPPORTED;
                    }
                    case SUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.4.PARTIALLY_SUPPORTED;
                    }
                    case INVALID: {
                        return EffectiveAlternativeSelector.AlternativeFitness.4.INVALID;
                    }
                    default: {
                        return EffectiveAlternativeSelector.AlternativeFitness.4.UNEVALUATED;
                    }
                }
            }
        }, 
        PARTIALLY_SUPPORTED {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                switch (assertionFitness) {
                    case UNKNOWN:
                    case UNSUPPORTED:
                    case SUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.5.PARTIALLY_SUPPORTED;
                    }
                    case INVALID: {
                        return EffectiveAlternativeSelector.AlternativeFitness.5.INVALID;
                    }
                    default: {
                        return EffectiveAlternativeSelector.AlternativeFitness.5.UNEVALUATED;
                    }
                }
            }
        }, 
        SUPPORTED_EMPTY {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                throw new UnsupportedOperationException("Combine operation was called unexpectedly on 'SUPPORTED_EMPTY' alternative fitness enumeration state.");
            }
        }, 
        SUPPORTED {
            @Override
            EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness assertionFitness) {
                switch (assertionFitness) {
                    case UNKNOWN:
                    case UNSUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.7.PARTIALLY_SUPPORTED;
                    }
                    case SUPPORTED: {
                        return EffectiveAlternativeSelector.AlternativeFitness.7.SUPPORTED;
                    }
                    case INVALID: {
                        return EffectiveAlternativeSelector.AlternativeFitness.7.INVALID;
                    }
                    default: {
                        return EffectiveAlternativeSelector.AlternativeFitness.7.UNEVALUATED;
                    }
                }
            }
        };
        
        abstract EffectiveAlternativeSelector.AlternativeFitness combine(PolicyAssertionValidator.Fitness p0);
    }
}
